#include "vbr_state.h"

// HCF v3.0
// Statistics utilities: look the bottom of
// this file for thier implemetations
long rnd32(long seed);
double uniform_0_1(long *seed);
double uniform(double a, double b, long *seed);
double negexp (double mean, long *seed);

VBR_state::VBR_state(int index_, long seed_)
{
   index = index_;
   seed = seed_;
   writeable = true;

   // default distributions
   interval_dist = INTERVAL_EXPO;
   pktsize_dist  = PKTSIZE_EXPO;
   // default parameters (mean bitrate 128kps)
   mean_interval = 0.01;
   mean_pktsize = 160;
}


VBR_state::~VBR_state()
{
}


bool VBR_state::addTransition(VBR_state* state, double prob)
{
   if (!writeable)
      return false;

   VBR_transition trans;
   trans.state = state;
   trans.prob = prob;
   transitions.push_back(trans);
   return true;
}


bool VBR_state::checkTransitions()
{
   bool am_here = false;
   double sum = 0.0;

   // getting the sum of all probabilities
   for (transition_it it = transitions.begin(); it != transitions.end(); ++it)
   {
      if ((*it).state == this)
         am_here = true;
      sum += (*it).prob;
   }

   // checks the sum and eventually adds a self-transition
   if (am_here)
      return (VBR_EQ_ONE(sum));
   else
   {
      if (VBR_GT_ONE(sum))
         return false;
      else if (VBR_LT_ONE(sum))
         addTransition(this, 1.0 - sum);
   }
   return true;
}


VBR_state* VBR_state::getNextState()
{
   //double guess = uniform_0_1(&seed);
   double guess = ((float)rand())/RAND_MAX;
   double sum = 0.0;

   // I want a real in [0.0, 1.0) (not so good, but what else we can do???)
   if (guess == 0.0) guess = VBR_DELTA;

   for (transition_it it = transitions.begin(); it != transitions.end(); ++it)
   {
      sum += (*it).prob;
      if (sum >= guess)
         return (*it).state;
   }
   // anyway we return the last state
   return (*(transitions.rbegin())).state;
}


int VBR_state::getPktSize()
{
   int next_pktsize;
   double guess;
   switch (pktsize_dist)
   {
      case PKTSIZE_EXPO:
         next_pktsize = (int)(negexp(mean_pktsize, &seed));
         if (next_pktsize > VBR_PKTSIZE_THRESHOLD)
            next_pktsize = VBR_PKTSIZE_THRESHOLD;
         return (next_pktsize);
      case PKTSIZE_TWOVAL:
         guess = uniform_0_1(&seed);
         return ((guess < 0.5) ? size1 : size2);
      case PKTSIZE_EQDIST:
         return (int)(uniform((double)size1, (double)size2, &seed));
   }
   fprintf(stderr, "ERROR: trying to use an undefined distribution while choosing packet size!\n");
   return -1;
}


double VBR_state::getInterval()
{
   switch (interval_dist)
   {
      case INTERVAL_EXPO:
         return (negexp(mean_interval, &seed));
         break;
      case INTERVAL_CONST:
         return (const_interval);
         break;
   }

   return -1.0;
}


int VBR_state::getMeanDataRate()
{
   return (int)(getAvgPktSize()*8/getAvgInterval());
}

double VBR_state::getAvgInterval()
{
   switch(interval_dist)
   {
      case INTERVAL_EXPO:
         return mean_interval;
      case INTERVAL_CONST:
         return const_interval;
      default:
         return -1.0;
   }
}


int VBR_state::getAvgPktSize()
{
   switch(pktsize_dist)
   {
      case PKTSIZE_EXPO:
         return mean_pktsize;
      case PKTSIZE_TWOVAL:
      case PKTSIZE_EQDIST:
         return (size1+size2)/2;
      default:
         return -1;
   }
}

/******************************************************
                  STATISTICS UTILITIES
******************************************************/

#include <math.h>

#define MODULE  2147483647
#define A       16807
#define LASTXN  127773
#define UPTOMOD -2836
#define RATIO   0.46566128e-9            /* 1/MODULE */


/*        
**  Function  : long rnd32(long seed) 
**  Return    : the updated value of 'seed'.
**  Remarks   : congruential generator of pseudorandom sequences of numbers
**              uniformly distributed between 1 and 2147483646, using the 
**              congruential relation: Xn+1 = 16807 * Xn  mod  2147483647 . 
**              The input to the routine is the integer Xn, while the returned 
**              integer is the number Xn+1.
*/
long rnd32(long seed)
{
   long times, rest, prod1, prod2;
   times = seed / LASTXN;
   rest  = seed - times * LASTXN;
   prod1 = times * UPTOMOD;
   prod2 = rest * A;
   seed  = prod1 + prod2;
   if(seed < 0) seed = seed + MODULE;
   return seed;
}


/*
**  Function  : double uniform_0_1(long *seed)
**  Return    : a value uniformly distributed between 0 and 1
**  Remarks   : the value of '*seed' is changed.
*/        
double uniform_0_1(long *seed)
{
   double u;
   *seed = rnd32 (*seed);
   u = (*seed) * RATIO;
   return u;
}


/*
**  Function  : double uniform(double a, double b, long *seed)
**  Return    : a value uniformly distributed between 'a' and 'b'
**  Remarks   : the value of '*seed' is changed.
*/        
double uniform(double a, double b, long *seed)
{
   double u;
   if (a > b)
   {
      u = a; a = b; b = u;
   }
   *seed = rnd32 (*seed);
   u = (*seed) * RATIO;
   u = a + u * (b-a);
   return u;
}


/*        
**  Function  : double negexp(double mean, long *seed)
**  Return    : a value exponentially distributed with mean 'mean'.
**  Remarks   : the value of '*seed' is changed.
*/        
double negexp (double mean, long *seed)
{
   double u;
   *seed = rnd32 (*seed);
   u = (*seed) * RATIO;
   return ( - mean*log(u) );
}
