/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) Xerox Corporation 1997. All rights reserved.
 *  
 * License is granted to copy, to use, and to make and to use derivative
 * works for research and evaluation purposes, provided that Xerox is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Xerox grants no other licenses expressed or implied. The Xerox trade
 * name should not be used in any advertising without its written permission.
 *  
 * XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
 * express or implied warranty of any kind.
 *  
 * These notices must be retained in any copies of any part of this software.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/trace/traffictrace.cc,v 1.14 2002/05/21 02:11:15 ddutta Exp $ (Xerox)";
#endif

/* XXX: have not dealt with errors.  e.g., if something fails during
 * TraceFile::setup(), or TrafficTrace is not pointing to a TraceFile,
 * no guarantee about results.
 */
 
#include <sys/types.h>
#include <sys/stat.h> 
#include <stdio.h>

#include "config.h"
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */


#include "random.h"
/* module to implement a traffic generator in ns driven by a trace
 * file.  records in the trace file consist of 2 32 bit fields, the
 * first indicating the inter-packet time in microseconds, and the
 * second indicating the packet length in bytes.  multiple TraffficTrace
 * objects can use the same trace file, and different TrafficTrace
 * objects can use different trace files.  For each TrafficTrace, a
 * random starting point within the trace file is selected.
 */

#include "object.h"
#include "trafgen.h"

struct tracerec {
        u_int32_t trec_time; /* inter-packet time (usec) */
	u_int32_t trec_size; /* size of packet (bytes */
};


/* object to hold a single trace file */

class TraceFile : public NsObject {
 public:
	TraceFile();
	void get_next(int&, struct tracerec&); /* called by TrafficGenerator
						* to get next record in trace.
						*/
	int setup();  /* initialize the trace file */
	int command(int argc, const char*const* argv);
 private:
	void recv(Packet*, Handler*); /* must be defined for NsObject */
        int status_; 
	char *name_;  /* name of the file in which the trace is stored */
	int nrec_;    /* number of records in the trace file */
	struct tracerec *trace_; /* array holding the trace */
};

/* instance of a traffic generator.  has a pointer to the TraceFile
 * object and implements the interval() function.
 */

class TrafficTrace : public TrafficGenerator {
 public:
	TrafficTrace();
	int command(int argc, const char*const* argv);
	virtual double next_interval(int &);
 protected:
	void timeout();
	TraceFile *tfile_;
	struct tracerec trec_;
	int ndx_;
	void init();
};


static class TraceFileClass : public TclClass {
 public:
	TraceFileClass() : TclClass("Tracefile") {}
	TclObject* create(int, const char*const*) {
		return (new TraceFile());
	}
} class_tracefile;

TraceFile::TraceFile() : status_(0)
{
}

int TraceFile::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "filename") == 0) {
			name_ = new char[strlen(argv[2])+1];
			strcpy(name_, argv[2]);
			return(TCL_OK);
		}
	}
	return (NsObject::command(argc, argv));
}

void TraceFile::get_next(int& ndx, struct tracerec& t)
{
	t.trec_time = trace_[ndx].trec_time;
	t.trec_size = trace_[ndx].trec_size;

	if (++ndx == nrec_)
		ndx = 0;

}

int TraceFile::setup()
{
	tracerec* t;
	struct stat buf;
	int i;
	FILE *fp;

	/* only open/read the file once (could be shared by multiple
	 * SourceModel's
	 */
	if (! status_) {
		status_ = 1;

		if (stat(name_, (struct stat *)&buf)) {
			printf("could not stat %s\n", name_);
			return -1;
		}

		nrec_ = buf.st_size/sizeof(tracerec);
		unsigned nrecplus = nrec_ * sizeof(tracerec);
		unsigned bufst = buf.st_size;

		//	if ((unsigned)(nrec_ * sizeof(tracerec)) != buf.st_size) {
		if (nrecplus != bufst) {
			printf("bad file size in %s\n", name_);
			return -1;
		}

		trace_ = new struct tracerec[nrec_];

		if ((fp = fopen(name_, "rb")) == NULL) {
			printf("can't open file %s\n", name_);
			return -1;
		}

		for (i = 0, t = trace_; i < nrec_; i++, t++)
			if (fread((char *)t, sizeof(tracerec), 1, fp) != 1) {
				printf("read failed\n");
				return -1 ;
			}
			else {
		
				t->trec_time = htonl(t->trec_time);
				t->trec_size = htonl(t->trec_size);
			}

	}

	/* pick a random starting place in the trace file */
	return (int(Random::uniform((double)nrec_)+.5));

}

void TraceFile::recv(Packet*, Handler*)
{
        /* shouldn't get here */
        abort();
}

/**************************************************************/

static class TrafficTraceClass : public TclClass {
 public:
	TrafficTraceClass() : TclClass("Application/Traffic/Trace") {}
	TclObject* create(int, const char*const*) {
	        return(new TrafficTrace());
	}
} class_traffictrace;

TrafficTrace::TrafficTrace()
{
	tfile_ = (TraceFile *)NULL;
}

void TrafficTrace::init()
{
	if (tfile_) 
		ndx_ = tfile_->setup();
}

int TrafficTrace::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	
	if (argc == 3) {
		if (strcmp(argv[1], "attach-tracefile") == 0) {
			tfile_ = (TraceFile *)TclObject::lookup(argv[2]);
			if (tfile_ == 0) {
				tcl.resultf("no such node %s", argv[2]);
				return(TCL_ERROR);
			}
			return(TCL_OK);
		}
	}

	return (TrafficGenerator::command(argc, argv));

}

void TrafficTrace::timeout()
{
        if (! running_)
                return;

        /* send a packet */
	// Note:  May need to set "NEW_BURST" flag in sendmsg() for 
	// signifying a new talkspurt when using vat traces.
	// (see expoo.cc, tcl/ex/test-rcvr.tcl)
	agent_->sendmsg(size_);
        /* figure out when to send the next one */
        nextPkttime_ = next_interval(size_);
        /* schedule it */
        timer_.resched(nextPkttime_);
}


double TrafficTrace::next_interval(int& size)
{
        tfile_->get_next(ndx_, trec_);
	size = trec_.trec_size;
	return(((double)trec_.trec_time)/1000000.0); /* usecs->secs */
}

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

/*
**  File name   : voicetrace.cc
**  Author      : Michele Garetto, Politecnico di Torino, Italy
**  Version     : 2.0
**  Purpose     : Module to implement a traffic generator in ns driven by a
**                trace file representing a bidirectional stream of VOIP packets.
**                Rows in the trace file correspond to single packets, and
**                contain the (relative) time (in microseconds) at which the
**                packet is sent and the RTP payload length (in bytes). The length
**                of talk spurts is taken from a Pareto distribution of mean
**		  paretomean and shape paretoshape, to be assigned from tcl.
**		  A single VoiceTraceFile is created for two VoiceTrace objects
**		  referring to the same VoiceTraceFile, identified by 1 or 2
**		  by a tcl command "id" (see examplevoice.tcl)
**
**		  This module is a modification of the file traffictrace.cc
**                belonging to the standard ns distribution
*/

#define CYCLICDELAY 20000  /* delay (microseconds) added while restarting the trace */

typedef struct voicerec {
        double time;      /* relative time (usec)   */
	int size;         /* size of packet (bytes) */
	struct voicerec* next; /* next packet in the trace */
} Trec;

/* object to hold a single trace file as a cyclic list of voicerec records */
class VoiceTraceFile : public NsObject {

 public:

	VoiceTraceFile() : status_(0), head_(0), tail_(0), curr_(0), fr_(0),
		call1_(0), call2_(0), debug_(0)
	{
		 bind("debug_", &debug_);
		 bind_time("paretomean_", &paretomean_);
	         bind("paretoshape_", &paretoshape_);
	}
	void setup();  /* initialize the trace file */
	double get_next(int id, int& size);
	double newspurt();
	int command(int argc, const char*const* argv);
	inline Trec* head() { return (head_); }
	inline Trec* tail() { return (tail_); }

 protected:

 	double paretomean_;  /* mean duration of talk spurts */
	double paretoshape_; /* shape parameter of Pareto distribution */

	void recv(Packet*, Handler*); /* must be defined for NsObject */
        int status_;
	char *name_;      /* name of the file in which the trace is stored */
	Trec* head_;      /* head of the list of packets */
	Trec* tail_;      /* tail of the list of packets */
	Trec* curr_;      /* next packet to send */
	int currcycles_;  /* current number of cycles since beginning of current talk spurt */

	Trec* fr_;        /* first record of a next talk spurt */
	int frcycles_;    /* number of cycles before first record of next talk spurt */

	int call1_,call2_;
	int debug_;
};

/* instance of a traffic generator.  has a pointer to the TraceFile
 * object and implements the interval() function.
 */

class VoiceTrace : public TrafficGenerator {

 public:

	VoiceTrace(): id_(0),tfile_(0), nextsize_(0), debug_(0)
	{
		bind("debug_", &debug_);
	}
	int command(int argc, const char*const* argv);
	virtual double next_interval(int &) { return 0.; }
	virtual void process_data(int size, AppData* data);

 protected:

	void timeout();
	int id_;
	VoiceTraceFile *tfile_;
	int nextsize_; /* size of next packet to send */

	int debug_;
	void init();
};


static class VoiceTraceFileClass : public TclClass {
 public:
	VoiceTraceFileClass() : TclClass("VoiceTracefile") {}
	TclObject* create(int, const char*const*) {
		return (new VoiceTraceFile());
	}
} class_voicetracefile;


int VoiceTraceFile::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "filename") == 0) {
			name_ = new char[strlen(argv[2])+1];
			strcpy(name_, argv[2]);
			return(TCL_OK);
		}
	}
	return (NsObject::command(argc, argv));
}

void VoiceTraceFile::setup()
{
	Trec* t;
	long num;

	struct stat buf;
	FILE *fp;
	char row[200];

	double time;
	int size;

	/* only open/read the file once (could be shared by multiple
	 * SourceModel's
	 */
	if (! status_) {

		status_ = 1;

		if (stat(name_, (struct stat *)&buf)) {
			printf("could not stat %s\n", name_);
			return;
		}

		if ((fp = fopen(name_, "r")) == NULL) {
			printf("can't open file %s\n", name_);
			return;
		}

		num = 0;

		while (fgets(row,200,fp) != NULL) {

			if (sscanf(row,"%lf %d",&time,&size) == 2) {
				t = new (struct voicerec);
				t->time = time;
				t->size = size;
				t->next = 0;

				if (!head_) head_ = tail_ = t;
    				else {
					tail_->next = t;
					tail_ = t;
				}

				num++;

				if(debug_>1) printf("Read packet %ld: time = %f - size = %d\n",num,time,size);
			}
		}

		fclose(fp);

		/* cyclic binding */
		if(head_) tail_->next = head_;

		/* first packet to send */
		curr_ = head_;

		if(debug_>1) printf("Duration samples (mean %f, shape %f):\n",paretomean_,paretoshape_);
		for(int i = 0; i<40; i++) {
			double dur = Random::pareto(paretomean_, paretoshape_);
			if(debug_>1) printf("dur = %f\n",dur);
		}

		if(debug_>1) {
			printf("Number of voice packets read = %ld\n",num);
		}
	}
}

void VoiceTraceFile::recv(Packet*, Handler*)
{
        /* shouldn't get here */
	printf("VoiceTraceFile::recv: shouldn't get here \n");
        exit(1);
}

/**************************************************************/

static class VoiceTraceClass : public TclClass {
 public:
	VoiceTraceClass() : TclClass("Application/Traffic/Trace/Voice") {}
	TclObject* create(int, const char*const*) {
	        return(new VoiceTrace());
	}
} class_voicetrace;

void VoiceTrace::init()
{
	if (tfile_) tfile_->setup();
	if (agent_)
		agent_->set_pkttype(PT_VOIP);
}

void VoiceTrace::process_data(int size, AppData* data)
{
	Scheduler& s = Scheduler::instance();
	if(debug_>1) printf("(%6.10f) VoiceTrace object %d: received %d bytes\n",s.clock(), id_, size);
}

int VoiceTrace::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();

	if (argc == 3) {
		if (strcmp(argv[1], "attach-tracefile") == 0) {
			tfile_ = (VoiceTraceFile *)TclObject::lookup(argv[2]);
			if (tfile_ == 0) {
				tcl.resultf("no such node %s", argv[2]);
				return(TCL_ERROR);
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "id") == 0) {

			id_ = atoi(argv[2]);
                        if(id_<1 || id_>2) {
				tcl.resultf("id %d not valid (must be 1 or 2)", id_);
				return(TCL_ERROR);
			}
			return (TCL_OK);
		}
	}

	return (TrafficGenerator::command(argc, argv));

}

void VoiceTrace::timeout()
{
	Scheduler& s = Scheduler::instance();

        if (! running_) return;

        /* send a packet */
	// Note:  May need to set "NEW_BURST" flag in sendmsg() for
	// signifying a new talkspurt when using vat traces.
	// (see expoo.cc, tcl/ex/test-rcvr.tcl)
	if(nextsize_ > 0) {
		agent_->sendmsg(nextsize_);
		if(debug_) {printf("(%6.10f) %5s VoiceTrace object %d: sending packet", s.clock(), name(), id_); getchar();}
	}

        /* figure out when to send the next one */
	double deltatime = tfile_->get_next(id_,nextsize_);
	if(debug_>1) printf("(%6.10f) VoiceTrace object %d: next packet after %f s\n", s.clock(), id_, deltatime);
	if(deltatime >= 0) timer_.resched(deltatime);
	else { printf("Error: deltatime < 0\n"); exit(1); }
}

double VoiceTraceFile::newspurt()
{
	Scheduler& s = Scheduler::instance();
	Trec* t;
	double dur, sw, tmp, sum;
	double deltatime;

	/* random duration of next talk spurt */
	dur = Random::pareto(paretomean_, paretoshape_);
	if(debug_>1) printf("(%6.10f) VoiceTraceFile:: duration of next talk spurt = %f sec \n",s.clock(),dur);
	dur*=1e6; /* secs -> usecs */

	/* theoretical switch instant */
	sw = curr_->time + dur;

	/* initialize currcycles */
	currcycles_ = 0;

	/* search first record of next talk spurt */
	tmp = 0.;
	sum = curr_->time;
	frcycles_ = 0;
	t = curr_;
	while(sum < sw)
	{
		t = t->next;
		if(t == head_) {
			tmp += tail_->time + CYCLICDELAY;
			frcycles_++;
		}
		sum = t->time + tmp;
	}
	fr_ = t;
	deltatime = dur + sum - sw;

	if(debug_>2) printf("(%6.10f) VoiceTraceFile:: first rec of new talk spurt will be: time %f - size %d\n",s.clock(),fr_->time, fr_->size);

	return(deltatime);
}

double VoiceTraceFile::get_next(int id, int& size)
{

	Scheduler& s = Scheduler::instance();
	double deltatime;

	if(!head_) return(-1);

	/* first calls to the procedure */
	if(id == 1 && !call1_) {
		deltatime = head_->time;
		size = head_->size;
		call1_ = 1;
		if(debug_) printf("(%6.10f) VoiceTraceFile: first call of speaker 1\n",s.clock());
		return(deltatime/1e6); /* usecs -> secs */
	}
	if(id == 2 && !call2_) {
		if(debug_) printf("(%6.10f) VoiceTraceFile: first call of speaker 2\n",s.clock());
		deltatime = newspurt();
		size = fr_->size;
		call2_ = 1;
		return(deltatime/1e6); /* usecs -> secs */
	}

	deltatime = -curr_->time;
	curr_ = curr_->next;
	deltatime += curr_->time;
	if(curr_ == head_) {
		currcycles_++;
		deltatime += tail_->time + CYCLICDELAY;
	}

	/* check if record is the first of the next talk spurt */
	if(curr_ == fr_ && currcycles_ == frcycles_) {
		deltatime += newspurt();
		size = fr_->size;
	} else {
		size = curr_->size;
	}

	return(deltatime/1e6); /* usecs -> secs */
}

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

